#include <stdio.h>
#include <stdlib.h>

//--------------------------------------------------------------------------
// Fortran column major indexing.  We don't actually need the number
// of columns (n), but we'll include it as an argument anyway for when
// we do C-style (row major) indexing.
#define IDX2F(i,j,m,n) (((j)*(m))+(i))

// C-style row major indexing.  We don't actually need the number
// of rows (m), but we'll include it as an argument anyway for when
// we do Fortran-style (column major) indexing.
//#define IDX2C(i,j,m,n) (((i)*(n))+(j))


//--------------------------------------------------------------------------
// Macro to help diagnose CUDA failures.
#define cudaTry(cudaStatus) _cudaTry(cudaStatus, __FILE__, __LINE__)

void _cudaTry(cudaError_t cudaStatus, const char *fileName, int lineNumber) {
  if(cudaStatus != cudaSuccess) {
    fprintf(stderr, "%s in %s line %d\n",
        cudaGetErrorString(cudaStatus), fileName, lineNumber);
    exit(1);
  }
}

//----------------------------------------------------------------------
void s418mm_cpu(float *A, float *B, float *C, uint m, uint n, uint p) {

  // Basic matrix multiplication code.  It will not be particularly
  // efficient since it was not coded in a cache aware fashion.
  for(uint i = 0; i < m; i++)
    for(uint k = 0; k < p; k++) {
      float sum = 0.0;
      for(uint j = 0; j < n; j++)
	sum += A[IDX2F(i,j,m,n)] * B[IDX2F(j,k,n,p)];
      C[IDX2F(i,k,m,p)] = sum;
    }
}


//----------------------------------------------------------------------
__global__ void s418mm_kernel(float *A, float *B, float *C,
			      uint m, uint n, uint p) {

  // Determine our location in the matrix using nvcc built-ins.
  uint i = blockIdx.x*blockDim.x + threadIdx.x;
  uint k = blockIdx.y*blockDim.y + threadIdx.y;

  // If we are not an extra thread.
  if((i < m) && (k < p)) {
    // Calculate a single element of the output matrix.
    float sum = 0.0;
    for(uint j = 0; j < n; j++)
      sum += A[IDX2F(i,j,m,n)] * B[IDX2F(j,k,n,p)];
    C[IDX2F(i,k,m,p)] = sum;
  }
}

//----------------------------------------------------------------------
void s418mm_gpu(float *A, float *B, float *C, uint m, uint n, uint p) {
  float *dev_A, *dev_B, *dev_C;
  int size_A = m * n * sizeof(float);
  int size_B = n * p * sizeof(float);
  int size_C = m * p * sizeof(float);

  // Allocate memory on the GPU for all of the arrays.
  cudaTry(cudaMalloc((void**)(&dev_A), size_A));
  cudaTry(cudaMalloc((void**)(&dev_B), size_B));
  cudaTry(cudaMalloc((void**)(&dev_C), size_C));
  
  // Copy over the input arrays.
  cudaTry(cudaMemcpy(dev_A, A, size_A, cudaMemcpyHostToDevice));
  cudaTry(cudaMemcpy(dev_B, B, size_B, cudaMemcpyHostToDevice));

  // Set up the computational grid.  We will use maximum size square
  // blocks, but your choice may have been different.
  dim3 dimGrid(ceil(m/32.0), ceil(p/32.0), 1);
  dim3 dimBlock(32, 32, 1);

  // Calculate the matrix multiplication.
  s418mm_kernel<<<dimGrid, dimBlock>>>(dev_A, dev_B, dev_C, m, n, p);

  // Copy back the result.
  cudaTry(cudaMemcpy(C, dev_C, size_C, cudaMemcpyDeviceToHost));

  // Free up the arrays on the GPU.
  cudaTry(cudaFree(dev_A));
  cudaTry(cudaFree(dev_B));
  cudaTry(cudaFree(dev_C));
}
