#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "timing418.h"
#include "hw4_kernel.h"

//--------------------------------------------------------------------------
// When comparing two single precision floating point numbers, what is
// close enough to be considered equality?
#define CLOSE_ENOUGH 1e-3

// Convenient allocation macro.
#define NewArray(Type, N) ((Type *)(malloc(N*sizeof(Type))))

// Fortran column major indexing.  We don't actually need the number
// of columns (n), but we'll include it as an argument anyway for when
// we do C-style (row major) indexing.
#define IDX2F(i,j,m,n) (((j)*(m))+(i))

// C-style row major indexing.  We don't actually need the number
// of rows (m), but we'll include it as an argument anyway for when
// we do Fortran-style (column major) indexing.
//#define IDX2C(i,j,m,n) (((i)*(n))+(j))

//--------------------------------------------------------------------------
// Generates a matrix with random entries between 0 and 1 (inclusive).
void rand_matrix(float *A, uint m, uint n) {
  // We use rows on the inner loop to be efficient in a column-major
  // indexing scheme.  Column major is the Fortran convention used by
  // BLAS (and is opposite the usual C row-major convention).
  for(uint j = 0; j < n; j++)
    for(uint i = 0; i < m; i++)
      A[IDX2F(i,j,m,n)] = (double)rand() / (double)RAND_MAX;
}

//----------------------------------------------------------------------
int compare_floats(const void *a,const void *b) {
  float *x = (float *) a;
  float *y = (float *) b;

  if (*x < *y) 
    return -1;
  else if (*x > *y) 
    return 1; 
  else
    return 0;
}

//----------------------------------------------------------------------
void s418mm_test(uint m, uint n, uint p, uint trials) {
  float *A, *B, *C_cpu, *C_gpu;
  timer_418 t0, t1;
  float *cpu_times, *gpu_times;

  // Make some space to store the results from the trials.
  cpu_times = NewArray(float, trials);
  gpu_times = NewArray(float, trials);

  // Allocate and fill the matrices.
  A = NewArray(float, m*n);
  rand_matrix(A, m, n);
  B = NewArray(float, n*p);
  rand_matrix(B, n, p);
  C_cpu = NewArray(float, m*p);
  C_gpu = NewArray(float, m*p);
  
  // Time the calculations.
  printf("CPU:");
  for(uint i = 0; i < trials; i++) {
    timestamp(&t0);
    s418mm_cpu(A, B, C_cpu, m, n, p);
    timestamp(&t1);
    cpu_times[i] = elapsed_seconds(t1, t0);
    printf(" %6.3e", cpu_times[i]);
  }
  printf("\nGPU:");
  for(uint i = 0; i < trials; i++) {
    timestamp(&t0);
    s418mm_gpu(A, B, C_gpu, m, n, p);
    timestamp(&t1);
    gpu_times[i] = elapsed_seconds(t1, t0);
    printf(" %6.3e", gpu_times[i]);
  }
  printf("\n");

  // Check the calculations.
  for(uint i = 0; i < m; i++)
    for(uint k = 0; k < p; k++)
      if(fabs(C_cpu[IDX2F(i,k,m,p)] - C_gpu[IDX2F(i,k,m,p)]) > CLOSE_ENOUGH) {
	fprintf(stderr, "ERROR: i=%d, k=%d, cpu=%f, gpu=%f\n",
		i, k, C_cpu[IDX2F(i,k,m,p)], C_gpu[IDX2F(i,k,m,p)]);
	exit(-1);
      }
  
  // Report the medians.
  qsort(cpu_times, trials, sizeof(float), compare_floats);
  qsort(gpu_times, trials, sizeof(float), compare_floats);
  printf("Medians: CPU %6.3e, GPU %6.3e, speedup %5.1e\n\n", 
	 cpu_times[trials/2], gpu_times[trials/2], 
	 cpu_times[trials/2] / gpu_times[trials/2]);

  free(cpu_times);
  free(gpu_times);
  free(A);
  free(B);
  free(C_cpu);
  free(C_gpu);
}


//----------------------------------------------------------------------
int main(int argc, char **argv) {

  char *conversion_garbage;
  uint m, n, p, trials, default_trials = 5;;

  if((argc < 2) || (argc > 5)) {
    fprintf(stderr, "usage: hw4 m [ n = m] [ p = n ] [ trials = %d ]\n", default_trials);
    exit(1);
  }

  // Parse the input arguments.
  m = strtoul(argv[1], &conversion_garbage, 10);
  if((m < 1) || (strlen(conversion_garbage) > 0)) {
    fprintf(stderr, "hw4, unable to convert %s to positive integer m", argv[1]);
    exit(1);
  }

  if(argc > 2) {
    n = strtoul(argv[2], &conversion_garbage, 10);
    if((n < 1) || (strlen(conversion_garbage) > 0)) {
      fprintf(stderr, "hw4, unable to convert %s to positive integer n", argv[2]);
      exit(1);
    }
  }
  else
    n = m;

  if(argc > 3) {
    p = strtoul(argv[3], &conversion_garbage, 10);
    if((p < 1) || (strlen(conversion_garbage) > 0)) {
      fprintf(stderr, "hw4, unable to convert %s to positive integer p", argv[3]);
      exit(1);
    }
  }
  else
    p = n;

  if(argc > 4) {
    trials = strtoul(argv[4], &conversion_garbage, 10);
    if((trials < 1) || (strlen(conversion_garbage) > 0)) {
      fprintf(stderr, "hw4, unable to convert %s to positive integer trials", argv[4]);
      exit(1);
    }
  }
  else
    trials = default_trials;

  // Uncomment the call to srand() if you want different random number
  // sequences on each run.
  //srand((uint)time(NULL));

  // Run the tests.
  printf("Running matrix mult with m = %d, n = %d, p = %d, trials = %d\n", 
	 m, n, p, trials);
  s418mm_test(m,n,p,trials);
}

